#define CSTACK_DEFNS 1
#define USE_RINTERNALS 1
#include <Rinternals.h>
#include <Rinterface.h> /* for C stack */

#include "threads.h"
#include <stdlib.h>
#include <stdio.h>  /* debugging */

/*---- implementation ---*/

#define TS_INIT   0  /* init state (thread is not running yet) */
#define TS_CREA   1  /* running state just after creation, still joined with R */
#define TS_RUN    2  /* running state in sync with R but after detaching */
#define TS_PAR    4  /* parallel state */
#define TS_ABORT  5  /* aborted state - waiting for sync after an error */
#define TS_FIN    6  /* finished state - waiting for sync after exit */

#include <pthread.h>

/* RCNTXT */
#include "cntxt.h"

#include "mutex.h"

static rt_mutex_t *R_main_mutex;

static void finalize_thread(SEXP thread) {
  thread_t *t = (thread_t*) EXTPTR_PTR(thread);
  printf("*> finalize %p\n", t);
  if (t->res)
    R_ReleaseObject(t->res);
  if (t->prot)
    R_ReleaseObject(t->prot);
  free(t);
}

/* this has been added in R 2.13.0 but is not officially part
   of the API since all this is experimental */
extern SEXP R_CollectFromIndex(PROTECT_INDEX i);

static void thread_enter_parallel_region(SEXP thread) {
  thread_t *t = (thread_t*) EXTPTR_PTR(thread);
  printf("-- entering parallel region\n");
  /* check whether we are already in the parallel region */
  if (t->state == TS_PAR) return;
  /* collect all objects that the thread put on the prot. stack */
  t->prot = R_CollectFromIndex(t->top);
  /* and make sure that object is protected */
  R_PreserveObject(t->prot);
  t->state = TS_PAR;
  mutex_unlock(R_main_mutex);
}

static void thread_leave_parallel_region(SEXP thread) {
  thread_t *t = (thread_t*) EXTPTR_PTR(thread);
  PROTECT_INDEX top;
  if (t->state != TS_PAR) return;
  printf("-- leaving parallel region, attempting sync\n");
  mutex_lock(R_main_mutex);
  printf("-- lock obtained, parallel region left\n");
  t->state = TS_RUN;
  R_ProtectWithIndex(t->self, &top);
  UNPROTECT(1);
  t->top = top;
  if (t->prot) { /* put protected objects back on the stack */
    SEXP p = t->prot;
    PROTECT_INDEX n = LENGTH(p);
    while (n--)
      PROTECT(VECTOR_ELT(p, n));
    R_ReleaseObject(p);
    t->prot = 0;
  }
}

static SEXP thread_return(SEXP thread, SEXP result) {
  thread_t *t = (thread_t*) EXTPTR_PTR(thread);
  printf("*> thread %p return (state=%d)\n", thread, t->state);
  if (result != R_NilValue)
    R_PreserveObject(result);
  t->res = result;
  if (t->state == TS_CREA || t->state == TS_RUN)
    R_CollectFromIndex(t->top);
  if (t->prot) {
    R_ReleaseObject(t->prot);
    t->prot = 0;
  }    
  t->state = TS_FIN;
  R_ReleaseObject(t->self);
  printf("-- thread finished\n");
  mutex_unlock(R_main_mutex);
  pthread_exit(t);
  return result;
}

static SEXP create_thread_object() {
  SEXP thread;
  thread_t *t = (thread_t*) malloc(sizeof(thread_t));
  if (!t) Rf_error("Cannot allocate thread descriptor");
  thread = R_MakeExternalPtr(t, R_NilValue, R_NilValue);
  R_PreserveObject(thread);
  R_RegisterCFinalizer(thread, finalize_thread);
  t->self = thread;
  t->enter_parallel_region = thread_enter_parallel_region;
  t->leave_parallel_region = thread_leave_parallel_region;
  t->call_return = thread_return;
  t->pt = 0;
  t->res = 0;
  t->prot = 0;
  t->state = TS_INIT;
  return thread;
}

/* if an error occurs in during the evaluation,
   clean up and close the thread */
static void thread_eval_cend(void *ptr) {
  thread_t *t = (thread_t*) ptr;
  printf("*> thread %p cend (state=%d)\n", ptr, t->state);
  if (t->prot) {
    R_ReleaseObject(t->prot);
    t->prot = 0;
  }    
  if (t->state == TS_CREA)
    R_CollectFromIndex(t->top);
  printf("-- thread aborted\n");
  t->state = TS_ABORT;
  R_ReleaseObject(t->self);
  mutex_unlock(R_main_mutex);
  pthread_exit(t);
}

static void *thread_eval_(void *ptr) {
  thread_t *t = (thread_t*) ptr;
  RCNTXT *cntxt = (RCNTXT*) t->rcx;
  printf("-- entered eval thread\n");
  t->state = TS_CREA;

  Rf_begincontext(cntxt, CTXT_CCODE, R_NilValue, R_BaseEnv, R_BaseEnv,
		  R_NilValue, R_NilValue);
  cntxt->cenddata = t;
  cntxt->cend = thread_eval_cend;
  
  printf("-- running eval\n");
  t->res = eval(t->expr, t->env);
  Rf_endcontext(cntxt);

  printf("-- eval returned cleanly (state=%d)\n", t->state);
  /* this should never happen (unless a package write makes a big
     mistake), but just in case sync with R */
  if (t->state == TS_PAR) {
    t->leave_parallel_region(t->self);
    Rf_warning("thread call ended in the parallel region!");
  }
  
  R_PreserveObject(t->res);
  /* remove all objects used by the thread - and drop it o
     the floor so it can be collected */
  R_CollectFromIndex(t->top);
  /* there should be no protected items, but just in case ... */
  if (t->prot) {
    R_ReleaseObject(t->prot);
    t->prot = 0;
  }
    
  t->state = TS_FIN;
  printf("-- thread finished\n");
  mutex_unlock(R_main_mutex);
  pthread_exit(t);
  return t;
}

SEXP yield() {
  mutex_unlock(R_main_mutex);
  mutex_lock(R_main_mutex);
  return R_NilValue;
}

SEXP thread_info(SEXP thread) {
  thread_t *t = (thread_t*) EXTPTR_PTR(thread);
  SEXP res = PROTECT(allocVector(VECSXP, 3));
  SET_VECTOR_ELT(res, 0, ScalarInteger(t->state));
  SET_VECTOR_ELT(res, 1, ScalarLogical(t->res ? 1 : 0));
  SET_VECTOR_ELT(res, 2, t->res ? t->res : R_NilValue);
  UNPROTECT(1);
  return res;
}

SEXP ptest(SEXP thread) {
  PROTECT(thread);
  printf("++ entering parallel region\n");
  R_enter_parallel_region(thread);
  printf("++ in parallel region\n");
  sleep(1);
  printf("++ leaving parallel region\n");
  R_leave_parallel_region(thread);
  PROTECT(thread);
  printf("++ back in sync, returning 2\n");
  UNPROTECT(2);
  R_thread_return(thread, Rf_ScalarInteger(2));
}

SEXP thread_eval(SEXP expr, SEXP env) {
  SEXP thread = create_thread_object();
  thread_t *t = (thread_t*) EXTPTR_PTR(thread);
  pthread_t *pt;
  pthread_attr_t ta;
  PROTECT_INDEX top;

  if (!R_main_mutex) {
    R_main_mutex = mutex_create();
    if (!R_main_mutex)
      Rf_error("unable to create global synchronization mutex");
    /* the mutex is always locked when R is busy */
    mutex_lock(R_main_mutex);
  }
  
  /* disable stack limit because we'll have separate stacks for threads */
  R_CStackLimit = (uintptr_t) -1;

  pt = (pthread_t*) malloc(sizeof(pthread_t));
  if (!pt)
    Rf_error("unable to allocate thread context");

  Rf_defineVar(install(".current.thread"), thread, R_GlobalEnv);

  /* this is really a trick to determine the current top of the stack */
  R_ProtectWithIndex(thread, &top);
  UNPROTECT(1);

  t->pt   = pt;
  t->expr = expr;
  t->env  = env;
  t->top  = top;
  pthread_attr_init(&ta);
  pthread_attr_setdetachstate(&ta, PTHREAD_CREATE_DETACHED);
  if (pthread_create(pt, &ta, thread_eval_, t) != 0) {
    free(pt);
    t->pt = 0;
    R_ReleaseObject(t->self);
    Rf_error("unable to create a new thread");
  }
  
  /* creating was successful, so it't time to wait to for the thread
     to finish or enter parallel region */
  printf("-> thread created, re-locking mutex\n");
  mutex_lock(R_main_mutex);
  printf("-> mutex locked, continuing on main thread\n");

  /* we land here a) on the first parallel region invocation or
     b) if the thread finished without even going parallel */

  /* remove the current thread as we're back on the main thread */
  Rf_defineVar(install(".current.thread"), R_NilValue, R_GlobalEnv);

  return thread;
}

SEXP e(SEXP x, SEXP r) {
  return eval(x, r);
}
